local T = CBW_EggRoboTag
local laser_idle = 0
local laser_charging = 1
local laser_firing = 2
local chargerate = FRACUNIT/TICRATE
local firetime = 10
local jumpthrust = FRACUNIT*5/4
local vlimit = 10
local hlimit = 16
local limitfactor = FRACUNIT*99/100
local chargerecoil = 16
local bombrings = 5
local bombcooldown = TICRATE/2
local lasercooldown = TICRATE/2
local burnlength = TICRATE*2

local EggLook = function(player)
	local twod = (twodlevel or player.mo.flags2 & MF2_TWOD)
	local ringdist, span
	if not(twod)
		ringdist = RING_DIST*2
		span = ANG30
	else
		ringdist = RING_DIST
		span = ANG20
	end

	local maxdist = FixedMul(ringdist, player.mo.scale)
	local closestdist = 0
	local closestmo = nil
	local nonenemiesdisregard = MF_SPRING
	searchBlockmap("objects",function(pmo,mo)
		if (mo.flags & MF_NOCLIPTHING) return end
		if (mo.health <= 0) return end -- dead

		if not(mo.player)
		and (!((mo.flags & (MF_ENEMY|MF_BOSS|MF_MONITOR)
		and (mo.flags & MF_SHOOTABLE)) or (mo.flags & MF_SPRING)) == !(mo.flags2 & MF2_INVERTAIMABLE)) -- allows if it has the flags desired XOR it has the invert aimable flag
			return -- not a valid target
		end
		//CTF monitor 
		if mo.type == MT_RING_REDBOX and not(G_GametypeHasTeams() and player.ctfteam == 1) then return end
		if mo.type == MT_RING_BLUEBOX and not(G_GametypeHasTeams() and player.ctfteam == 2) then return end

		if (mo == pmo) return end
		if (mo.flags2 & MF2_FRET) return end
		if (mo.flags & nonenemiesdisregard) return end
		if mo.type == MT_EGGROBOTAGBOMB return end //This is eggrobo's projectile!
		if (mo.player and ((G_TagGametype() and mo.player.pflags&PF_TAGIT == player.pflags&PF_TAGIT) or mo.player.spectator)) then return end //Disallow targeting teammates


		//Do angle/distance checks
		local zdist = (pmo.z + pmo.height/2) - (mo.z + mo.height/2)
		local dist = P_AproxDistance(pmo.x-mo.x, pmo.y-mo.y)
		//CBW: 	Made the angle checks their own locals, for readability purposes
		//		I also unsigned the angle checks, which appears to correct failed OutOfBounds checks above the player.
		local xyz_angle = abs(R_PointToAngle2(0, 0, dist, zdist))
		local xy_angle = abs(R_PointToAngle2(
				pmo.x + P_ReturnThrustX(pmo, pmo.angle, pmo.radius),
				pmo.y + P_ReturnThrustY(pmo, pmo.angle, pmo.radius),
				mo.x, mo.y
			) - pmo.angle)
			
			
		dist = P_AproxDistance(dist, zdist)
		if (dist > maxdist)
			return -- out of range
		end
-- 		print("\x82!!!VERT","# Span check: "..xyz_angle/ANG1,"# Limit: "..span*2/ANG1,"#Within Range: "..tostring(abs(xyz_angle) <= span))
		if (xyz_angle > span)
			return -- Don't home outside of desired angle!
		end


		if (twod
		and abs(pmo.y-mo.y) > pmo.radius)
			return -- not in your 2d plane
		end

		if ((closestmo and closestmo.valid) and (dist > closestdist))
			return
		end
-- 		print("\x82!!!HORZ","# Span check: "..xy_angle/ANG1,"# Limit: "..span*2/ANG1,"#Within Range: "..tostring(abs(xy_angle) <= span))
		if (xy_angle > span)
			return -- behind back
		end

		if not (P_CheckSight(pmo, mo))
			return -- out of sight
		end

		closestmo = mo
		closestdist = dist
	end,player.mo,player.mo.x-maxdist,player.mo.x+maxdist,player.mo.y-maxdist,player.mo.y+maxdist)
	return closestmo
end

T.EggReset = function(player)
	//Egg Robo stats
	player.iseggrobo = false
	player.ert_state = 0
	player.ert_tics = 0
	player.ert_cooldown = 0
	player.ert_aimy = 0
	player.ert_charged = 0
	//Skin stats
	local skin = skins[player.realmo.skin]
	player.charability = skin.ability
	player.charability2 = skin.ability2
	player.charflags = skin.flags
	player.normalspeed = skin.normalspeed
	player.acceleration = skin.acceleration
	player.accelstart = skin.accelstart
	player.thrustfactor = skin.thrustfactor
	player.followitem = skin.followitem	
	player.shieldscale = skin.shieldscale
	player.camerascale = skin.camerascale
	player.height = skin.height
end

T.EggPreThink = function(player)
	if player.mo and player.mo.valid and player.iseggrobo and player.ert_state == laser_firing
		player.cmd.angleturn = player.mo.angle>>FRACBITS
	end
end

//Player Mechanics, general input
T.EggPlayerThink = function(player)
	local mo = player.mo
	if not(mo and mo.valid) then return end
	//Initialize variables
	if player.iseggrobo == nil
		player.iseggrobo = false
	end
	if player.ert_state == nil
		player.ert_state = 0
	end
	if player.ert_tics == nil
		player.ert_tics = 0
	end
	if player.ert_cooldown == nil
		player.ert_cooldown = 0
	end
	if player.ert_buttonhistory == nil
		player.ert_buttonhistory = player.cmd.buttons
	end
	if player.ert_aimy == nil
		player.ert_aimy = 0
	end
	if player.ert_charged == nil
		player.ert_charged = 0
	end
	
	//Gates
	if not(player.iseggrobo) then return end
	if P_PlayerInPain(player)
	or player.powers[pw_nocontrol]
	or player.playerstate == PST_DEAD
	or player.ert_state == -1
		player.ert_state = -1
		player.ert_tics = 0
		player.ert_cooldown = 0
		return
	end
	
	//Stat adjustment
	local refskin = skins[0]
	player.normalspeed = refskin.normalspeed
	player.charability = CA_GLIDEANDCLIMB
	player.charability2 = 0
	player.height = 80*FRACUNIT
	player.acceleration = refskin.acceleration
	player.accelstart = refskin.accelstart
	player.thrustfactor = refskin.thrustfactor
	player.shieldscale = FRACUNIT*3/2
	if not(P_IsObjectOnGround(mo))
		player.thrustfactor = $<<1
	end
	
	local do_hlimit,do_vlimit = false,false
	local bombbt = BT_ATTACK
	local doaction = (player.cmd.buttons&bombbt) and not(player.ert_buttonhistory&bombbt)
	//Vertical movement
	if player.cmd.buttons&BT_JUMP
		//Do thrust
		if not(mo.eflags&MFE_UNDERWATER)
			P_SetObjectMomZ(mo,jumpthrust,true)
		else
			P_SetObjectMomZ(mo,jumpthrust*2/3,true)			
		end
	end
	
	//Bomb
	if doaction 
		if not(player.ert_cooldown) and player.rings >= bombrings then 
			local missile = P_SPMAngle(mo,MT_EGGROBOTAGBOMB,mo.angle,0)
			if missile and missile.valid then
				missile.momx = 0
				missile.momy = 0
				S_StartSound(mo,sfx_cdfm47)
				S_StartSound(mo,sfx_cdfm66,player)
				player.rings = $-bombrings
				player.ert_cooldown = bombcooldown
			end
		else
			S_StartSound(mo,sfx_s3k8c,player)
		end
	end
	
	//Laser targeting
	if player.ert_state == laser_idle or not(mo.target and mo.target.valid)
		mo.target = EggLook(player)
	end
	local t = mo.target
	
	if t and t.valid and not(P_CheckSight(mo, t))
		t = nil
	end
	
	//Laser target display
	if t != nil and t.valid
		P_SpawnLockOn(player, t, mobjinfo[MT_LOCKON].spawnstate)
	end
	
	//Laser charging
	if player.ert_state == laser_charging
		do_vlimit = true
		do_hlimit = true
		
		//Focus sprite forward
		if t and t.valid
			mo.angle = R_PointToAngle2(mo.x,mo.y,t.x,t.y)
		end
		player.drawangle = mo.angle
		
		//Increase charge amount
		player.ert_charged = min(FRACUNIT,$+chargerate)
		
		//Get release
		if not(player.cmd.buttons&BT_SPIN)
			player.ert_state = laser_firing
			player.ert_tics = 0
			player.ert_aimy = -ANG1*5
			S_StartSound(mo,sfx_s3k5e)
			if player.ert_charged == FRACUNIT
				P_Thrust(mo,mo.angle+ANGLE_180,chargerecoil*mo.scale)
				mo.momz = $/4
				T.DebugPrint("Fully charged blast!",player)
			else
				P_Thrust(mo,mo.angle+ANGLE_180,chargerecoil*mo.scale>>1)
				mo.momz = $/2
				T.DebugPrint("Eggrobo laser fire started",player)
			end
		end
	end


	
	//Laser firing
	if player.ert_state == laser_firing
		player.drawangle = mo.angle
		do_vlimit = true
		do_hlimit = true
		if t and t.valid
			local dist = R_PointToDist2(mo.x,mo.y,t.x,t.y)
			local zdist = (t.z)-(mo.z+mo.height/2)
			player.ert_aimy = R_PointToAngle2(0,0,dist,zdist)
			if P_MobjFlip(mo) == -1 then
				player.ert_aimy = -$
			end
			mo.angle = R_PointToAngle2(mo.x,mo.y,t.x,t.y)
-- 			mo.angle = R_PointToAngle2(mo.x-FixedMul(cos(ANGLE_45+$),FRACUNIT*),
		end
		local aiming = player.aiming
		player.aiming = player.ert_aimy
		local projectile = MT_EGGROBOTAGSHOT
		if player.ert_charged == FRACUNIT
			projectile = MT_EGGROBOTAGSHOT2
		end
		local e = P_SPMAngle(mo,projectile,mo.angle,1)
		if e and e.valid 
			if projectile == MT_EGGROBOTAGSHOT 
				e.scale = $<<1
			else
				e.scale = $>>1
			end
		end
		if e and e.valid then
			T.ERLShift(e)
		end
		if player.ert_tics >= firetime
			player.ert_aimy = 0
			player.ert_state = laser_idle
			player.ert_tics = -lasercooldown
			T.DebugPrint("Eggrobo laser fire ended",player)
		end
		player.aiming = aiming
	end
	
	//Air drag
	if not(P_IsObjectOnGround(mo))
		local factor = FRACUNIT * 95 / 100
		local hspeed = FixedMul(factor,FixedHypot(mo.momx,mo.momy))
		local fullspeed = FixedHypot(hspeed,mo.momz)
		local angle3d = R_PointToAngle2(0,0,hspeed,mo.momz)
		local angle2d = R_PointToAngle2(0,0,mo.momx,mo.momy)
		hspeed = FixedMul(fullspeed,cos(angle3d))
		local vspeed = FixedMul(fullspeed,sin(angle3d))
		mo.momx = FixedMul(hspeed,cos(angle2d))
		mo.momy = FixedMul(hspeed,sin(angle2d))
		mo.momz = vspeed
	end
	//Enforce speed limits
	local factor = FixedDiv(mo.friction,limitfactor)
	local speed = FixedHypot(player.rmomx,player.rmomy)
	local angle = R_PointToAngle2(0,0,player.rmomx,player.rmomy)
	if do_hlimit
	and speed > hlimit*mo.scale
		P_InstaThrust(mo,angle,FixedMul(speed,factor))
		mo.momx = $+player.cmomx
		mo.momy = $+player.cmomy
	end
	if do_vlimit
	and abs(mo.momz) > vlimit*mo.scale
		mo.momz = FixedMul($,factor)
	end
	
		
	player.ert_cooldown = max(0,$-1)
	player.ert_tics = $+1
end

//Player Input
T.EggJumpControl = function(player)
	if not(player.iseggrobo) then return end
	if not(player.iseggrobo)
	or player.powers[pw_nocontrol]
	or player.ert_state == laser_firing
	return true end
	if not(player.ert_buttonhistory&BT_JUMP)
		S_StartSound(player.mo,sfx_s3ka0)
-- 		T.DebugPrint("Eggrobo propulsion started",player)
	end
	return true
end

T.EggSpinControl = function(player)
	if not(player.iseggrobo) then return end
	if not(player.ert_state == laser_idle and player.ert_tics >= 0)
	or player.powers[pw_nocontrol]
	return true end
	if not(player.ert_buttonhistory&BT_SPIN)
		player.ert_state = laser_charging
		player.ert_tics = 0
		player.ert_charged = 0
		//Start laser charge GFX
		local charge = P_SpawnMobjFromMobj(player.mo,0,0,0,MT_EGGROBOTAGCHARGE)
		charge.target = player.mo
		T.DebugPrint("Eggrobo laser charge started",player)
	end
	return true
end

//Projectile Mechanics
T.BombSphereTouch = function(mo,playermo)
	if mo.target == playermo
	or playermo.player.iseggrobo
	return true end
	T.DebugPrint("Bomb sphere touch")
	P_DamageMobj(playermo,mo,mo.target)
	P_KillMobj(mo)
	return true
end

T.BombSphereMissileCollide = function(mo,collide)
	if mo.z < collide.z+collide.height
	and mo.z+mo.height > collide.z
	and collide.flags&MF_MISSILE then
		T.DebugPrint("Bomb sphere missile collision")
		P_KillMobj(mo)
		P_KillMobj(collide)
		S_StartSound(mo,sfx_s1c4)
	return true end
end

T.ERLShift=function(mo)
	if mo.target and mo.target.valid and mo.target.player and mo.target.player.ert_state != 0 then
		local dist = mo.target.scale*76
		local angle = -ANGLE_22h
		angle = $+mo.target.angle
		local x = mo.target.x+P_ReturnThrustX(nil,angle,dist)
		local y = mo.target.y+P_ReturnThrustY(nil,angle,dist)
		local z = mo.target.z+mo.target.scale*48
		P_TeleportMove(mo,x,y,z)
	else
		P_RemoveMobj(mo)
	end
end

T.Afterburn = function(mo)
	if P_IsObjectOnGround(mo)
		local p = P_SpawnMobjFromMobj(mo,0,0,0,MT_SPINFIRE)
		p.target = mo.target
		p.fuse = burnlength
		p.state = S_EGGROBOFIRE1
		p.scale = $<<3
		if not(G_GametypeHasTeams())
			p.color = SKINCOLOR_CYAN
		elseif p.target and p.target.valid and p.target.player
			if p.target.player.ctfteam == 1
				p.color = SKINCOLOR_RED
			end
			if p.target.player.ctfteam == 2
				p.color = SKINCOLOR_BLUE
			end
		end
	end
end